//
//	bsendm
//			Author Tatsuo Baba
//			2000.03.08 porting win32 -> linux
//			last update 2000.03.16
//			last update 2000.06.14  file '|'
//			last update 2000.10.28  interval time for inetd
//			last update 2000.11.02  qmail-queue
/*
command line mailer

  bsendm [-g] [-s] [-t] [-v] [-e] parameter-file

  parameter file: must be euc-code if kanji is used
1: [>]log-file-name	[tab]	notify mail address
2: [helo-host/]smtp-ip-address[:port][:time-out] [tab] interval-time for inetd 
3: mailto or ${environment}
4: mailfrom
5: subject
6: attachment-file-1 [tab] attachment-file-2 ...
7: body

-s : substitute mode  %1

-t : test mode

-v : verbose mode

-g : print progress '.'

-e : log-file exclusive mode

  parameter file example:
#-------------- start ----------
/dev/null
127.0.0.1 or /var/qmail/bin/qmail-queue
${SENDER}   .... qmail-command
webmaster@babaq.net
subject-1

body-start
bye.


	
  command example:
$ bsendm ./mail.txt
$ bsendm -s ./mfile

*/

#include <unistd.h>
#include <sys/stat.h>
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <bsmtp.h>




class CStr
{
public:
	CStr(){m_pData = NULL;m_nSize=m_nCur=0;};
	~CStr(){if (m_pData) delete []m_pData;};
	void Clear(){m_nCur=0;};
	void Detach(){m_pData = NULL;};
	void Add(char *data,int len);
	void AddByte(int data);
	int GetLength(){return m_nCur;};
	char *GetBuffer(){return m_pData;};
private:
	char *m_pData;	// buffer pointer
	int m_nSize;	// buffer size
	int m_nCur;		// current length
};

void CStr::Add(char *data,int len)
{
	if (len <= 0)
		return;
	if (m_nSize <= m_nCur + len ) {
		char *tp = new char[m_nSize+len+2048];
		if (tp == NULL) {
			if (m_pData)
				delete [] m_pData;
			throw "CStr out of space";
		}
		if (m_pData) {
			memcpy(tp,m_pData,m_nCur);
			delete [] m_pData;
		}
		m_pData = tp;
		m_nSize = m_nSize + len + 2048;
	}
	memcpy(m_pData+m_nCur,data,len);
	m_nCur += len;
	m_pData[m_nCur] = 0;
}

void CStr::AddByte(int data)
{
	if (m_nSize <= m_nCur + 1 ) {
		char *tp = new char[m_nSize+2048];
		if (tp == NULL) {
			if (m_pData)
				delete [] m_pData;
			throw "CStr out of space";
		}
		if (m_pData) {
			memcpy(tp,m_pData,m_nCur);
			delete [] m_pData;
		}
		m_pData = tp;
		m_nSize += 2048;
	}
	m_pData[m_nCur] = data & 0x00ff;
	m_nCur++;
	m_pData[m_nCur] = 0;
}

class BLock
{
public:
	BLock(){m_handle = -1;}
    int Lock(const char* fname,int tries){
        if (m_handle == -1) {
			strcpy(m_fname,fname);
			strcat(m_fname,"_LOCK_FILE");
            while (tries-- > 0) {
                m_handle = open(m_fname,O_RDWR | O_CREAT | O_EXCL,0444);
                if (m_handle != -1)
                    break;
                sleep(1);
            }
            if (m_handle == -1) {
                unlink(m_fname);  // may be safe
            }
        }
        return m_handle;
    }
    void Unlock(){ if (m_handle != -1) {close(m_handle);
                    unlink(m_fname);m_handle=-1;};}
    ~BLock(){ Unlock();}
private:
    int m_handle;         // handle for lock
    char m_fname[256];    // file name
};


void putlog(char *msg);
void putlog2(char *msg,char *tofield);
void puttestlog(char *msg);			// 1999.10.24 test log
int deploy_env(char* src,char* dest);

int getline(FILE *fp,char *buff,int size);
int getlinex(FILE *fp,char *buff,int size);
int getline2(FILE *fp,char *buff,int size);
void sendmail(CStr& body);
void sendnotifymail();	// 1999.11.09 support notify mail


int DoProcessFile(const char* pfile);

int gtestmode = 0;	// 1999.10.24 supprt test mode
int gverbosemode = 0;	// 2000.03.08 supprt verbose mode
int gexclusive = 0;		// 2000.03.10 exclusive logfile
unsigned int interval = 2;	// 2000.10.29 interval time for inetd
int progress = 0;	// 2000.11.02 progress bar

int gcancel = 0;


int gsubstitute = 0;
int testsendcnt = 0; // 1999.10.27

FILE *testfp = NULL;

#define MAXLINE 1024
#define MAXSIZE 1024*20

char buffer[MAXSIZE],serverip[MAXLINE],to[MAXSIZE],from[MAXLINE];
char subj[MAXLINE],files[MAXLINE],logfile[MAXLINE],tofile[MAXLINE];
char svrpara[MAXLINE];
// 1999.11.09 Support Notify mail
char *notifyid = NULL;
int notify_ctr = 0;
int notify_errctr = 0;


int len,ctr;
int ok;
char msg[1024];


void main(int argc,char **argv)
{

    if (argc < 2) {
		printf("arg error\n");
		exit(1);
	}
	char *tp;
	int i;
	for (i = 1;i < argc; i++) {
		tp = argv[i];
		if (tp[0] != '-')
			break;
		//  '-' found
		tp++;
		while (tp[0] > ' ') {
			switch (tp[0]) {
			case 's':
				gsubstitute = 1; // set substitute flag
				break;
			case 't':
				gtestmode = 1;
				break;
			case 'v':
				gverbosemode = 1;
				break;
			case 'e':
				gexclusive = 1;
				break;
			case 'g':
				progress = 1;
				break;
			default:
				printf("invalid arg %s\n",tp);
				exit(1);
			}
			tp++;
		}

	}
	char *fname = tp;		// file name
	if (fname == NULL || fname[0] == 0) {
		printf("arg error\n");
		exit(1);
	}

	return DoProcessFile(fname);

}


int DoProcessFile(const char* pfile)
{

	FILE *infp1; 
	if ((infp1 = fopen(pfile, "r")) == NULL) {
		printf("cannot open %s\n",pfile);
		exit(1);
	}
	
	notifyid = NULL;	// 1999.11.09 Support notify
	getlinex(infp1,logfile,sizeof(logfile));	// logfile name
	if (logfile[0] != '\0') {
		notifyid = strchr(logfile,'\t');	// 1999.11.09 Support notify
		if (notifyid) {
			notifyid[0] = 0;
			notifyid++;
		}
		notify_ctr = 0;
		notify_errctr = 0;
	}
	getlinex(infp1,serverip,sizeof(serverip));	// serverip ip address
	if (serverip[0] == '\0') {
		fclose(infp1);
		putlog("parameter error - serverip ip address");
		return 1;
	}
	if (strstr(serverip,"qmail-queue"))
		interval = 0;
	char *tp = strchr(serverip,'\t'); // 2000.10.28 support interval-time
 	if (tp) {
		tp[0] = 0;
		interval = atol(tp+1);
	}
	strcpy(svrpara,"-euc ");	// encoding type
	strcat(svrpara,serverip);
	getlinex(infp1,to,sizeof(to));	// to
	if (to[0] == '\0') {
		fclose(infp1);
		putlog("parameter error - invalid mailto");
		return 0;
	}
	// 2000.03.09 check ${env} in mail address and replace it
	int rc = deploy_env(to,buffer);
	if (rc < 0) { // error
		fclose(infp1);
		putlog("parameter error - environment not enclosed");
		return 0;
	}
	strcpy(to,buffer);

	getlinex(infp1,from,sizeof(from));	// from
	if (from[0] == '\0') {
		fclose(infp1);
		putlog("parameter error - invalid mailfrom");
		return 0;
	}
	getlinex(infp1,subj,sizeof(subj));	// subject

	
	getlinex(infp1,files,sizeof(files));// file name[s]

	struct stat buf;
	int lastpos = 0;
	char *tpx;
	if (*files != '\0') {
		char *tp = new char[strlen(files) +2];
		memcpy(tp,files,strlen(files) +1);
		for (register int i = 0;tp[i] != 0;i++){
			if (tp[i] == '\t') {	// find tab 
				tp[i] = 0;
				tpx = strchr(tp+lastpos,'|');//2000.06.14
				if (tpx)
					tpx[0] = 0;
				if (0 != stat( tp + lastpos, &buf ))
					goto file_notfound;
				lastpos = i + 1;
			}
		}
		tpx = strchr(tp+lastpos,'|');//2000.06.14
		if (tpx)
			tpx[0] = 0;
	
		if (0 != stat( tp + lastpos, &buf )) {
			file_notfound:
			char temp[4096];
			fclose(infp1);
			strcpy(temp,"parameter error - file not found ");
			strcat(temp,tp+lastpos);
			putlog(temp);
			delete [] tp;
			return 0;
		}
		delete [] tp;
	}

	// read entire message body
	CStr body,newbody;
	while ( len = getline2(infp1,buffer,sizeof(buffer)) ) 
		body.Add(buffer,len);

	if (infp1 != NULL) fclose(infp1);	/* close File */

	
	if ((to[0] == '.' || to[0] == '/') &&  // file name specfy ?
		(infp1 = fopen(to,"r")) != NULL	) { // mail address file ?
		strcpy(tofile,to);
		while ( len = getlinex(infp1,to,sizeof(to)) && !gcancel)
			sendmail(body);
		if (infp1 != NULL) fclose(infp1);	/* close File */

		sendnotifymail();	// 1999.11.09 Support Notify mail

		return 0;
	}
	sendmail(body);

	sendnotifymail();	// 1999.11.09 Support Notify mail

	return 0;

}



void sendmail(CStr& body)
{

    time_t ttStart,ttStop,ttDiff;
    // get the start time
    ttStart=time(NULL);

	char* bodyp = body.GetBuffer() ;
	CStr newbody,repdata;

	// 1999.08.19 to size over 1050 cause error
	char toshort[512],*tofield;
	int tolen = strlen(to);
	if (tolen < 256) {
		strcpy(toshort,to);
		tofield = NULL;
	} else {
		strcpy(toshort," ..... ");
		tofield = to;
	}


	if (gsubstitute) {
		char *tp1 = strchr(to,'\t'); // substitute character 1
		if (tp1 == NULL) {
			sprintf(buffer,"none substitute character %s %s %s %s",serverip,toshort,from,subj);
			putlog2(buffer,tofield);
			goto skip;
		}
		tp1[0] = 0;
		tp1++;
		char *tpfile = strchr(tp1,'\t'); // substitue file
		if (tpfile != NULL) {
			FILE *infpx;
			tpfile[0] = 0;
			tpfile++;
			
			if ((infpx = fopen(tpfile, "r")) == NULL) {
				sprintf(buffer,"none substitute file %s %s %s %s",serverip,toshort,from,subj);
				putlog2(buffer,tofield);
				return;
			}
			char buffer[4097]; int len;
			while ( len = getline2(infpx,buffer,sizeof(buffer)) ) 
				repdata.Add(buffer,len);

			fclose(infpx);
			// binary check
			char *tx = repdata.GetBuffer();
			char *endp = tx + repdata.GetLength();
			
			for (char ch = 0;ch < 9;ch++) {
				char *p = strchr(tx,ch);
				if (p && strchr(tx,ch) < endp) {
					sprintf(buffer,"substitute error - binary file %s %s %s %s",serverip,toshort,from,subj);
					putlog2(buffer,tofield);
					return;
				}
			}

		}

		char *tp = body.GetBuffer();
		int len = body.GetLength();
		char ch1,ch2;
		for (int i = 0;i < len; i++) {
			ch1 = tp[i];
			ch2 = tp[i+1];
			if (ch1 == '%' && ch2 == '1' 
						&& ( i == 0 || tp[i-1] != '%')) {
				newbody.Add(tp1,strlen(tp1));
				i++;
				sprintf(buffer,"substitute  %s",tp1);
				putlog(buffer);
				continue;
			}
			if (ch1 == '%' && ch2 == '2'  
						&& ( i == 0 || tp[i-1] != '%')) {
				newbody.Add(repdata.GetBuffer(),repdata.GetLength());
				i++;
				sprintf(buffer,"substitute file name  %s",tpfile);
				putlog(buffer);
				continue;
			}
			newbody.AddByte(ch1);
		}
		bodyp = newbody.GetBuffer();
	}
skip:



	if (gcancel)	// 1999.10.22 check cancel
		return;
	
	if (gtestmode) {	// 1999.10.24 testmode start
		testsendcnt++;
		char temp[128];
		sprintf(temp,"====> %6ld contents ===  To: ",testsendcnt);
		puttestlog(temp);
		puttestlog(to);
		puttestlog("\nBody:\n");
		puttestlog(bodyp);		// 2000.02.22 FIX testmode log invalid
		puttestlog("\nFile:\n");
		puttestlog(files);
		puttestlog("\n");

		return;
	}

	// put a start log
	sprintf(buffer,"start send %s %s %s %s",serverip,toshort,from,subj);
	putlog2(buffer,tofield);
	// start sending
	
	ok = BSendMail(svrpara,   // SMTP server to negotiate with
                to,       // to mail address (e.g., who@who.com\twho2...)
                from,     // The e-mail address as you wish it to appear to the recipient
                subj,  // Message subject
                bodyp,     // Message body
                files,msg);    // Attach Files  split by tab char

	if (progress)
		printf(".");
	
	if (!ok) {
		notify_errctr++;	// 1999.11.09 Support notify
		sprintf(buffer,"send error (%s) %s %s %s %s",msg,serverip,toshort,from,subj);
		putlog2(buffer,tofield);
		return;
	}
	notify_ctr++;	// 1999.11.09 Support notify
	sprintf(buffer,"end   send %s %s %s %s",serverip,toshort,from,subj);
	putlog2(buffer,tofield);
	if (interval) { // 2000.10.29 interval specified ?
		sleep(interval);
	}
	return;
}


/*************************************************************************/
/* put log */
void putlog(char *msg)
{
	
	if (gtestmode)
		return;
	
	char buf[10240],ftime[256];
	time_t timer = time(NULL);	// <time.h>
	struct tm *lct = localtime(&timer);
	strftime(ftime,255,"%Y/%m/%d %X",lct);
	sprintf(buf,"%s  %s\n",ftime,msg);
	
	if (logfile[0] == '>') {	// 1999.10.22 clear mode
		strcpy(logfile,logfile+1);
		remove(logfile);
	}

	if (gverbosemode)
		printf(buf);

	if (strcmp(logfile,"/dev/null") == 0)
		return;

	BLock ex;
	if (gexclusive) {	// need lock
		ex.Lock(logfile,10);
	}

	FILE *fp = fopen(logfile,"a");	
	if (fp == NULL) {
		return;
	}

	fwrite(buf,1,strlen(buf),fp);


	fclose(fp);
	return;

}

// 1999.08.19 fix to many to field cause error  
void putlog2(char *msg,char *tofield)
{
	
	if (gtestmode)
		return;
	char buf[10240],ftime[256];
	time_t timer = time(NULL);	// <time.h>
	struct tm *lct = localtime(&timer);
	strftime(ftime,255,"%Y/%m/%d %X",lct);
	sprintf(buf,"%s  %s\n",ftime,msg);
	
	if (logfile[0] == '>') {	// 1999.10.22 clear mode 
		strcpy(logfile,logfile+1);
		remove(logfile);
	}

	if (gverbosemode)
		printf(buf);

	if (strcmp(logfile,"/dev/null") == 0)
		return;

	BLock ex;
	if (gexclusive) {	// need lock
		ex.Lock(logfile,10);
	}

	FILE *fp = fopen(logfile,"a");	
	if (fp == NULL) {
		return;
	}

	fwrite(buf,1,strlen(buf),fp);

	// 1999.08.19 to field is long, write 1 line
	if (tofield && *tofield != 0) {
		char *tp = tofield;
		char sv;
		int ctr = 1;
		while (tp) {
			char *tp2 = strchr(tp,'\t');
			if (tp2) {
				sv = tp2[0];
				tp2[0] = 0;
			}
			sprintf(buf,"\t%d\t%s\n",ctr++,tp);
			fwrite(buf,1,strlen(buf),fp);
			if (tp2) {
				tp = tp2 +1;
				tp2[0] = sv;
			} else
				tp = NULL;
		}
	}

	fclose(fp);
	return;

}


/* get line ignorring comment line */
int getline(FILE *fp,char *buff,int size)
{
	while (1) {
		if (fgets(buff,size,fp) == NULL) {
			buff[0] = '\0';
			return 0;
		}
		if (buff[0] != '#')
			break;
	}
	return strlen(buff);
}

/* read line with ignore comment */
int getlinex(FILE *fp,char *buff,int size)
{
	int len = getline(fp,buff,size);
	if (len == 0)
		return 0;
	if (buff[len-1] == '\n') {		// erase lf
		buff[len-1] = '\0';
		len--;
	}
	return len;
}



/*  read 1 line with crlf  */
int getline2(FILE *fp,char *buff,int size)
{
	if (fgets(buff,size,fp) == NULL) {
		buff[0] = '\0';
		return 0;
	}
	// 1999.07.08 Fix put CR/LF
	int len = strlen(buff);
	if (buff[len-1] != '\n')	// 1999.07.16 fix not cr
		return len;
	memcpy(buff+len-1,"\r\n\0",3);
	return len + 1;
}



/* put test log */
void puttestlog(char *msg)
{
	if (testfp == NULL) {
		if (logfile[0] == '>') {	// 1999.10.22 clear mode
			strcpy(logfile,logfile+1);
			remove(logfile);
		}
		testfp = fopen(logfile,"a");	
		if (testfp == NULL) {
			return;
		}
	}
	fwrite(msg,1,strlen(msg),testfp);
	return;
}



void sendnotifymail() {	// 1999.11.09 Support Notify mail
	
	if (!notifyid || gcancel || gtestmode)
		return;

	
	char nsubject[256],msg[80];
	time_t timer = time(NULL);	// <time.h>
	struct tm *lct = localtime(&timer);
	strftime(nsubject,255,"bsendm send complete %Y/%m/%d %X",lct);

	char buff[2048];
	sprintf(buff,"%s\r\n %ld sent   %ld error(s) \r\nlogfile: %s\r\n",
		subj,notify_ctr + notify_errctr,notify_errctr,logfile);

	int ok = BSendMail(svrpara,   // SMTP server to negotiate with
                notifyid,       // to mail address (e.g., who@who.com\twho2...)
                from,     // The e-mail address as you wish it to appear to the recipient
                nsubject,  // Message subject
                buff,     // Message body
                "",		// no attach file 
				msg);    // Attach Files  split by tab char

	return;

}

// deploy environment variable  ${xxx}
int deploy_env(char* src,char* dest)
{
	char *tp = strchr(src,'$');	
	if (!tp) {
		strcpy(dest,src);
		return 0;
	}
	int ctr = 0;
	dest[0] = 0;
	char *top = src;
	while (tp) {
		if (tp[1] == '{') {
			tp[0] = 0;
			strcat(dest,top);
			char *last= strchr(tp+1,'}');
			if (last == NULL)
				return -1;
			last[0] = 0;
			char *env = getenv(tp+2);
			if (env == NULL)
				return -1;
			ctr++;	
			strcat(dest,env);
			tp = last + 1;
			top = tp;
		} else
			tp++;
		tp = strchr(tp,'$');
	}
	strcat(dest,top);
	return ctr;
}
