//
//	bwsock.cc
//				Author Tatsuo Baba
//
#include <unistd.h>
#include "bwsock.h"
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <stdio.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/select.h>

#define  CR  '\015'
#define  LF  '\012'

#define  INVALID_SOCKET -1
#define  INVALID_HANDLE_VALUE -1

extern int errno;

int error_timeout =
#ifdef ETIMEDOUT
ETIMEDOUT;
#else
-7;
#endif

int error_inprogress =
#ifdef EINPROGRESS
EINPROGRESS;
#else
-8;
#endif

int error_wouldblock =
#ifdef EWOULDBLOCK
EWOULDBLOCK;
#else
-9;
#endif

// 2000.10.29 for debug
void PutLog(char* str)
{
        FILE *fp = fopen("./absmtp.log","a");
        if (fp == NULL) {
                return;
        }
        fwrite(str,1,strlen(str),fp);
        fclose(fp);
        return;
}
class BTrace
{
public:
	BTrace(char *str){
		strcpy(m_name,str);
		PutLog(m_name); 
		PutLog(" start\n"); }
	~BTrace(){
		PutLog(m_name);
		PutLog(" end\n"); }
private:
	char m_name[512];
};


BSocket::BSocket()
{
	m_bClosed = 0;
	m_destSocket = INVALID_SOCKET;
}

BSocket::~BSocket()
{

	if (m_destSocket != INVALID_SOCKET)
		Close();
}

int BSocket::Open()
{
	m_bClosed = 0;		// closed flag
    return m_destSocket = socket(AF_INET,SOCK_STREAM,0);

}

void BSocket::Close()
{
	if (m_destSocket == INVALID_SOCKET)
		return;
	int nRet;
	char szBuf[4096];
	m_bClosed = 1;
	m_RcvPos = 0;
	m_RcvData.Clear();
	shutdown(m_destSocket, 1);
	while(1)
	{
		nRet = recv(
				m_destSocket,	// socket
				szBuf,			// data buffer
				sizeof(szBuf),	// length of buffer
				0);				// recv flags
		if (nRet == 0 || nRet == -1)  
			break;
	}
	shutdown(m_destSocket, 2);
	// Close the socket
	close(m_destSocket);
	m_destSocket = INVALID_SOCKET;
}

// return code 0 : OK
//        -1 : ng 
//        -2 : timeout 
int BSocket::Connect(char* ipaddr,int port,int timeout)
{
	if (m_destSocket == INVALID_SOCKET)
		Open();

	m_port = port;
	m_timeout = timeout;

	if (m_destSocket == INVALID_SOCKET)
		return -1;

	unsigned long ip = inet_addr(ipaddr);
	memcpy(&m_destSockAddr.sin_addr,&ip,sizeof(ip));
	m_destSockAddr.sin_port = htons(port);
	m_destSockAddr.sin_family = AF_INET;
	
	// set delay mode
        if (fcntl(m_destSocket,F_SETFL,fcntl(m_destSocket,F_GETFL,0) | O_NONBLOCK) == -1)
		return -1;

	m_RcvPos = 0;
	if (connect(m_destSocket,(struct sockaddr*)&m_destSockAddr,
		sizeof(m_destSockAddr)) == 0) {
		fcntl(m_destSocket,F_SETFL,fcntl(m_destSocket,F_GETFL,0) & ~O_NONBLOCK);
                return 0;
	}
//	if ((errno != error_inprogress) && (errno != error_wouldblock)) return -1;
	fd_set wfds;
	struct timeval tv;
        FD_ZERO(&wfds);
	FD_SET(m_destSocket,&wfds);
	tv.tv_sec = m_timeout;
	tv.tv_usec = 0;			// wait for socket ready 
	// 2000.10.29 fix detect timeout 
	int rc = select(m_destSocket + 1,(fd_set*)0,&wfds,(fd_set*)0,&tv);
	if (rc == 0) { // timeout ?
		errno = error_timeout;
		return -2;
	}
	if (rc == -1)
		return -1;
	if (FD_ISSET(m_destSocket,&wfds)) {	// ok
		unsigned int dum = sizeof(m_destSockAddr);	
		char ch;	// get remote name
		if (getpeername(m_destSocket,(struct sockaddr*)&m_destSockAddr,&dum) == -1 ) {
			read(m_destSocket,&ch,1);
			return -1;
		}
		fcntl(m_destSocket,F_SETFL,fcntl(m_destSocket,F_GETFL,0) & ~O_NONBLOCK);
		return 0;
	}
	errno = error_timeout;
	return -2;
}		
	

// return code : 0 OK
//		-1 Closed
//		-2 Timeout
int BSocket::WriteString(char* buf,int len)
{
	if (m_bClosed)
		return -1;
	fd_set wfds;
	struct timeval tv;
	tv.tv_sec = m_timeout;
	tv.tv_usec = 0;
	FD_ZERO(&wfds);
	FD_SET(m_destSocket,&wfds);

	// 2000.10.29 fix detect timeout 
	int rc = select(m_destSocket + 1,(fd_set*)0,&wfds,(fd_set*)0,&tv);
	if (rc == 0) { // timeout ?
		errno = error_timeout;
		return -2;
	}
	if (rc == -1)
		return -1;
	if (FD_ISSET(m_destSocket,&wfds)) {	// ok
		return write(m_destSocket,buf,len);
	}
	errno = error_timeout;
	return -2;
}

static char* GetLine(BCStr &data, int *pos)
{
	int len = data.GetLength() - *pos;
	char *tp = data.GetBuffer() + *pos;
	for (int i = 0;i < len;i++) {
		if ( tp[i] == LF || (tp[i] == CR && (i + 1) < len
			&& tp[i+1] == LF)  ) {
			*pos += i + 2;
			if (tp[i] == LF)
				*pos -= 1;
			tp[i] = 0;
			if (data.GetLength() <= *pos) {
				*pos = 0;
				data.Clear();
			}
			return tp;
		}
	}
	return NULL;
}

// return code 0 : ok
//	-1 Closed
//	-2 Timeout
//      -3 NO DATA
int BSocket::ReadString(char* buf,int blen)
{
	buf[0] = 0;
	if (m_bClosed)
		return -1;
	if (m_RcvData.GetLength() > 0 &&
		m_RcvData.GetLength() > m_RcvPos) {
		char *tp = GetLine(m_RcvData,&m_RcvPos);
		if (tp) {
			if (blen < strlen(tp))
				tp[blen -1] = 0;
			strcpy(buf,tp);
			return 0;
		}
	}
	if (m_RcvPos > 4096) {
		m_RcvData.Compact(m_RcvPos);
		m_RcvPos = 0;
	}
	char bufs[8192];
	fd_set rfds;
	struct timeval tv;
	int loopdetect = 0; // 2000.10.29 FIX loop
rcvloop:
	tv.tv_sec = m_timeout;
	tv.tv_usec = 0;
	FD_ZERO(&rfds);
	FD_SET(m_destSocket,&rfds);

	// 2000.10.29 fix detect timeout 
	int rc = select(m_destSocket + 1,&rfds,(fd_set*)0,(fd_set*)0,&tv);
	if (rc == 0) { // timeout ?
		errno = error_timeout;
		return -2;
	}
	if (rc == -1)
		return -1;
	if (FD_ISSET(m_destSocket,&rfds)) {	// ok
		int rlen = read(m_destSocket,bufs,8192);
		m_RcvData.Add(bufs,rlen);
		if (m_RcvData.GetLength() > 0 && 
			m_RcvData.GetLength() > m_RcvPos) {
			char *tp = GetLine(m_RcvData,&m_RcvPos);
			if (tp) {
				if (blen < strlen(tp))
					tp[blen -1] = 0;
				strcpy(buf,tp);
				return 0;
			}
                }
		if (loopdetect++ > 1000) {
			errno = error_timeout;
			return -2;
		}
		goto rcvloop;	// waiting for cr lf
	}
	errno = error_timeout;
	return -2;
}
