//
// qminter.cc
//
#include <unistd.h>
#include <signal.h>
#include <time.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/wait.h>
#include "qminter.h"
#include "bclass.h"

#define OPT_DELETE_HEADER 1
#define OPT_BCC_FROM_FILE 2
#define OPT_TRANSFER 4
#define OPT_TEST_MODE 8


void get_msgid(char*from,char *msgid);
void GetCurrentRFC822Date(char * pszBuffer);

static int error_intr =
#ifdef EINTR
EINTR;
#else
-1;
#endif


extern int errno;


#define SUBSTDIO_OUTSIZE 8192
#define wait_crashed(w) ((w) & 127)
#define wait_exitcode(w) ((w) >> 8)

typedef struct substdio {
  char *x;
  int p;
  int n;
  int fd;
} substdio;


struct qmail {
  int flagerr;
  unsigned long pid;
  int fdm;
  int fde;
  substdio ss;
  char buf[1024];
} ;


static int fd_copy(int to,int from)
{
  if (to == from) return 0;
  if (fcntl(from,F_GETFL,0) == -1) return -1;
  close(to);
  if (fcntl(from,F_DUPFD,to) == -1) return -1;
  return 0;
}

static int fd_move(int to,int from)
{
  if (to == from) return 0;
  if (fd_copy(to,from) == -1) return -1;
  close(from);
  return 0;
}

static void substdio_fdbuf( register substdio *s, register int fd, 
			register char *buf, register int len)
{
  s->x = buf;
  s->fd = fd;
  s->p = 0;
  s->n = len;
}

static int allwrite(register int fd,register char *buf, register int len)
{
  register int w;
  while (len) {
    w = write(fd,buf,len);
    if (w == -1) {
      if (errno == error_intr) continue;
      return -1; /* note that some data may have been written */
    }
    if (w == 0) ; /* luser's fault */
    buf += w;
    len -= w;
  }
  return 0;
}

static int substdio_flush(register substdio *s)
{
  register int p;
  p = s->p;
  if (!p) return 0;
  s->p = 0;
  return allwrite(s->fd,s->x,p);
}

static int substdio_put(register substdio *s, register char *buf, register int len)
{
  register int n;
  n = s->n;
  if (len > n - s->p) {
    if (substdio_flush(s) == -1) return -1;
    /* now s->p == 0 */
    if (n < SUBSTDIO_OUTSIZE) n = SUBSTDIO_OUTSIZE;
    while (len > s->n) {
      if (n > len) n = len;
      if (allwrite(s->fd,buf,n) == -1) return -1;
      buf += n;
      len -= n;
    }
  }
  /* now len <= s->n - s->p */
  //byte_copy(s->x + s->p,len,buf);
  memcpy(s->x + s->p,buf,len);
  s->p += len;
  return 0;
}

static int substdio_puts(register substdio *s, register char *buf)
{
  return substdio_put(s,buf,strlen(buf));
}


static int wait_pid(int *wstat, int pid)
{
  int r = waitpid(pid,wstat,0);
  while ((r == -1) && (errno == error_intr)) 
    r = waitpid(pid,wstat,0);
  return r;
}

static int qmail_open(struct qmail *qq,char* qmailqueue)
{
  int pim[2];
  int pie[2];

  if (pipe(pim) == -1) return -1;
  if (pipe(pie) == -1) { close(pim[0]); close(pim[1]); return -1; }
 
  switch(qq->pid = fork()) {
    case -1:
      close(pim[0]); close(pim[1]);
      close(pie[0]); close(pie[1]);
      return -1;
    case 0:
      close(pim[1]);
      close(pie[1]);
      if (fd_move(0,pim[0]) == -1) _exit(120);
      if (fd_move(1,pie[0]) == -1) _exit(120);
      //if (chdir(auto_qmail) == -1) _exit(61);
      char *binqqargs[2];
      binqqargs[0] = qmailqueue;
      binqqargs[1] = 0;
      execv(*binqqargs,binqqargs);
      _exit(120);
  }

  qq->fdm = pim[1]; close(pim[0]);
  qq->fde = pie[1]; close(pie[0]);
  substdio_fdbuf(&qq->ss,qq->fdm,qq->buf,sizeof(qq->buf));
  qq->flagerr = 0;
  return 0;
}

static unsigned long qmail_qp(struct qmail *qq)
{
  return qq->pid;
}

static void qmail_fail(struct qmail *qq)
{
  qq->flagerr = 1;
}

static void qmail_put(struct qmail *qq, char *s, int len)
{
  if (!qq->flagerr) if (substdio_put(&qq->ss,s,len) == -1) qq->flagerr = 1;
}

static void qmail_puts(struct qmail *qq, char *s)
{
  if (!qq->flagerr) if (substdio_puts(&qq->ss,s) == -1) qq->flagerr = 1;
}

static void qmail_from(struct qmail *qq, char *s)
{
  if (substdio_flush(&qq->ss) == -1) qq->flagerr = 1;
  close(qq->fdm);
  substdio_fdbuf(&qq->ss,qq->fde,qq->buf,sizeof(qq->buf));
  qmail_put(qq,"F",1);
  if (s[0] == '<')
    qmail_put(qq,s+1,strlen(s+1)-1);
  else
    qmail_puts(qq,s);
  qmail_put(qq,"",1);
}

static void qmail_to(struct qmail *qq, char *s)
{
  qmail_put(qq,"T",1);
  if (s[0] == '<')
    qmail_put(qq,s+1,strlen(s+1)-1);
  else
    qmail_puts(qq,s);
  qmail_put(qq,"",1);
}

static char *qmail_close(struct qmail *qq)
{
  int wstat;
  int exitcode;

  qmail_put(qq,"",1);
  if (!qq->flagerr) if (substdio_flush(&qq->ss) == -1) qq->flagerr = 1;
  close(qq->fde);

  if (wait_pid(&wstat,qq->pid) != qq->pid)
    return "Zqq waitpid surprise (#4.3.0)";
  if (wait_crashed(wstat))
    return "Zqq crashed (#4.3.0)";
  exitcode = wait_exitcode(wstat);

  switch(exitcode) {
    case 115: /* compatibility */
    case 11: return "Denvelope address too long for qq (#5.1.3)";
    case 31: return "Dmail server permanently rejected message (#5.3.0)";
    case 51: return "Zqq out of memory (#4.3.0)";
    case 52: return "Zqq timeout (#4.3.0)";
    case 53: return "Zqq write error or disk full (#4.3.0)";
    case 0: if (!qq->flagerr) return ""; /* fall through */
    case 54: return "Zqq read error (#4.3.0)";
    case 55: return "Zqq unable to read configuration (#4.3.0)";
    case 56: return "Zqq trouble making network connection (#4.3.0)";
    case 61: return "Zqq trouble in home directory (#4.3.0)";
    case 63:
    case 64:
    case 65:
    case 66:
    case 62: return "Zqq trouble creating files in queue (#4.3.0)";
    case 71: return "Zmail server temporarily rejected message (#4.3.0)";
    case 72: return "Zconnection to mail server timed out (#4.4.1)";
    case 73: return "Zconnection to mail server rejected (#4.4.1)";
    case 74: return "Zcommunication with mail server failed (#4.4.2)";
    case 91: /* fall through */
    case 81: return "Zqq internal bug (#4.3.0)";
    case 120: return "Zunable to exec qq (#4.3.0)";
    default:
      if ((exitcode >= 11) && (exitcode <= 40))
	return "Dqq permanent problem (#5.3.0)";
      return "Zqq temporary problem (#4.3.0)";
  }
}


int MailToqmail(char* szServer,char* szTo[], int uToArrSize,
     int *nToFlag,char* szFrom,char* szSubject, char* szData,char* szFile,
                                                  int option,char *msg)
{
  // szServer is qmail-queue full path to invoke
  char from[256],msgid[128],work[512];
  int rc,i;
  char *qqx,*tp,*mailfrom;
  struct qmail qq;
  BCStr tostr,ccstr,replystr,mailto,header;

  int special_header = 0;
  msg[0] = 0;
  signal(SIGPIPE,SIG_IGN); // ignore pipe write failer

  mailfrom = strrchr(szFrom,'<');
  if (!mailfrom) {      // support   xxx@xxx.com (xxxxx) format
    strcpy(from,szFrom);
    mailfrom = strrchr(from,'(');
    if (mailfrom) {
      *mailfrom = 0;
      while (mailfrom[-1] == ' ') {  // supress blank
        mailfrom[-1] = 0; mailfrom--;
      }
    }
    mailfrom = from;
  }
  // create Message=ID
  get_msgid(mailfrom,msgid);

  for (i = 0; i < uToArrSize; i++) {
    switch( nToFlag[i] ) {
      case 1:
        tostr.Add(",\n  ",4);
        tostr.Add((char*)szTo[i],strlen(szTo[i]));
	break;
      case 2:
	ccstr.Add(",\n  ",4);
	ccstr.Add((char*)szTo[i],strlen(szTo[i]));
	break;
      case 3:
	break;
      case 4:
	replystr.Add((char*)szTo[i],strlen(szTo[i]));
        continue;
      case 5:
	special_header++ ;
	continue;
      //2000.01.14 Support TO: @header
      case 6:
	tostr.Add(",\4  ",4);
	tostr.Add((char*)szTo[i],strlen(szTo[i]));
	continue;		// not send as RCPT TO
    }		
    tp = strrchr(szTo[i],'<');
    if (!tp) {	// support   xxx@xxx.com (xxxxx) format
      strcpy(work,(char*)szTo[i]);
      tp = strrchr(work,'(');
      if (tp) {
	*tp = 0;
	while (tp[-1] == ' ') {  // supress blank
          tp[-1] = 0; tp--;
	}
      }
      tp = work;
    }
    mailto.Add(tp,strlen(tp));
    mailto.Add("",1);
  }

  // 2000.01.21 Support bcc from file
  if (option & OPT_BCC_FROM_FILE) {
    char* bccfile = 0;
    for (int i = 0; i < uToArrSize; i++) {
      if (strncasecmp(szTo[i],">#bcc ",6) == 0) {
	bccfile = szTo[i] + 6;
	break;
      }
    }
    BFile ls(bccfile);
    char *line;
    int lno = 0;
    while (line = ls.GetLine()){
      lno++;
      if (line[0] == '#' || line[0] == 0)
        continue;
      if (line[0] != '<') {
        mailto.Add("<",1);
        mailto.Add(line,strlen(line));
        mailto.Add(">",1);
      } else {
	mailto.Add(line,strlen(line));
      }
      mailto.Add("",1);
    }
  }

  if (tostr.GetLength() > 0) {
    header.Add("To: ",4);
    header.Add(tostr.GetBuffer()+4,tostr.GetLength()-4);
    header.Add("\n",1);
  }
  if (ccstr.GetLength() > 0) {
    header.Add("Cc: ",4);
    header.Add(ccstr.GetBuffer()+4,ccstr.GetLength()-4);
    header.Add("\n",1);
  }
  if (replystr.GetLength() > 0) {
    header.Add("Reply-To: ",10);
    header.Add(replystr.GetBuffer(),replystr.GetLength());
    header.Add("\n",1);
  }

  // Send special Header
  if (special_header > 0) {
    for (int i = 0; i < uToArrSize; i++) {
      if (nToFlag[i] != 5 || szTo[i] == NULL)
	continue;
      tp = (char*)szTo[i];
      if (tp[0] != '>')
	continue;
      if (tp[1] == '#') {
	continue;
      }
      // 2000.09.14 FIX multiple content-type
      if (strncasecmp(tp+1,"content-type",12) == 0)
         continue;
      if (strncasecmp(tp+1,"message-id:",10) == 0)
	msgid[0] = 0;
      header.Add(tp+1,strlen(tp+1));
      header.Add("\n",1);
    }
  }

  if (msgid[0] != 0) {
    header.Add(msgid,strlen(msgid));
  }

  int notransfer = !(option & OPT_TRANSFER);
  if (notransfer) { // need date header ?
    // send date header 
    char datebuf[128];
    GetCurrentRFC822Date(datebuf);
    header.Add(datebuf,strlen(datebuf));
    header.Add("From: ",6);
    header.Add((char*)szFrom,strlen(szFrom));
    header.Add("\n",1);
  }

  if (notransfer && szSubject) {
    header.Add("Subject: ",9);
    header.Add((char*)szSubject,strlen(szSubject));
    header.Add("\n",1);
  }

  char newfile[L_tmpnam];
  if (NULL == tmpnam(newfile))
    newfile[0] = 0;
  FILE *hfile = fopen(newfile,"wb");
  if (hfile == NULL){
    strcpy(msg,"cannot create qmail tempfile");
    return 0;
  }
  fwrite(header.GetBuffer(),1,header.GetLength(),hfile);

  if (szData) {
    fwrite((char*)szData,1,strlen(szData),hfile);
  }
  fclose(hfile);

  // now we are ready to invloke qmail-queue
  rc = qmail_open(&qq,szServer);
  if (rc != 0) { // qmail-queue invoke error
    sprintf(msg,"qmail open error %d",rc);
    return 0;
  }
  {
    BFile ls(newfile);
    char *line;
    while (line = ls.GetLine()){
      qmail_puts(&qq,line);
      qmail_puts(&qq,"\n");
    }
  }
  remove(newfile); // delete work file
  if (szFile) {
    BFile ls(szFile);
    char *line;
    while (line = ls.GetLine()){
      qmail_puts(&qq,line);
      qmail_puts(&qq,"\n");
    }
  }
  qmail_from(&qq,mailfrom);
  tp = mailto.GetBuffer();
  while (*tp) {
    qmail_to(&qq,tp);
    tp = tp + strlen(tp) + 1;
  }
  qqx = qmail_close(&qq);
  if (*qqx) {
    sprintf(msg,"qmail close error %s\n",qqx);
    return 0;
  }
  return 1;
}
