//
//  rfc1468.cc
//					Author Tatsuo Baba
//
//  SMTP Non-ASCII character support
//	ISO-2022-JP  => JIS KANJI support
//	BIG5   => Taiwan code support  1999.08.26
//	X-Mail ver 1.0 -> 1.1          2000.11.03
//
//
#include <stdlib.h>
#include "rfc1468.h"
#include "cbase64.h"
#include "bclass.h"
#include "kconv.h"

#define CRLF "\015\012"


static char TOJIS[] = "\033$B";
static char TOASCII[] = "\033(B";

static char MIME_HEADER[] = 
"MIME-Version: 1.0\015\012"
"X-Mail-Agent: bsmtp-1.1  "
__DATE__
"  by Tatsuo Baba\015\012";
static char MIME_BODY_JP[] = 
"Content-Type: text/plain; charset=ISO-2022-JP\015\012\015\012";

// 1999.08.26 Support other charset 2000.09.14 fix content-type
static char MIME_BODY_OTHER[] = 
"Content-Type: %s; charset=%s\015\012"
"Content-Transfer-Encoding: base64\015\012\015\012";

static char MIME_BODY[] = 
"Content-Type: text/plain\015\012\015\012";

static char MIME_BODY_WITHCONTENT[] = 
"Content-Type: %s; charset=%s\015\012\015\012";


static char MIME_BODY_MULTI_JP[] = 
"Content-Type: multipart/mixed; boundary=\"%s\"\015\012"
"\015\012--%s\015\012"
"Content-Type: text/plain; charset=ISO-2022-JP\015\012"
"Content-Transfer-Encoding: 7bit\015\012\015\012";




static char MIME_BODY_MULTI[] = 
"Content-Type: multipart/mixed; boundary=\"%s\"\015\012"
"\015\012--%s\015\012"
"Content-Type: text/plain\015\012"
"Content-Transfer-Encoding: 7bit\015\012\015\012";

static char MIME_BODY_MULTI2[] = 
"Content-Type: multipart/mixed; boundary=\"%s\"\015\012"
"\015\012--%s\015\012"
"%s\015\012"
"Content-Transfer-Encoding: %s\015\012\015\012";


static char MIME_BOUNDARY_END[] = 
"\015\012--%s--\015\012";

// 2000.05.17 Support varying Content-type
static char MIME_FILEPART[] = 
"\015\012--%s\015\012"
"Content-Type: %s;\015\012"
" name=\"%s\"\015\012"
"Content-Transfer-Encoding: base64\015\012\015\012";

static char MIME_FILEPART_7BIT[] = 
"\015\012--%s\015\012"
"Content-Type: %s;\015\012"
" name=\"%s\"\015\012"
"Content-Transfer-Encoding: 7bit\015\012\015\012";



static void ConvOneDot(BCStr &in,BCStr &out)
{
	out.Clear();
	int len = in.GetLength();
	char *tp = in.GetBuffer();
	for (int i = 0;i < len;i++) {
//		if ((tp[i] == '\n' || tp[i] == '\012') && tp[i+1] == '.' ) {
		if (tp[i] == '\n' && tp[i+1] == '.' && (i+1) < len) {
			out.Add(tp+i,2);
			out.Add(".",1);
			i += 1;
			continue;
		}
		out.Add(tp+i,1);
	}
}

/*
	Function: RFC1468::ConvSubject
	Description:
		Convert Subject.
*/
char *CRFC1468::ConvSubject(const char* in)
{
	BCStr out;
	char jisbuf[1024],buffer[1024];
	int clen;
	int len = Sjis2Jis(in,&out);
	if (len == 0) {			// no dbcs ? 
		char* temp = out.GetBuffer();
		out.Detach();
		return temp;		// no conversion
	}

	char* temp = out.GetBuffer();

	clen = jisleft(temp,36,jisbuf);	// get 36 bytes
	m_encode.EncodeLine(strlen(jisbuf),jisbuf,buffer,sizeof(buffer));
	BCStr str;
	int mlen = strlen(m_strMIME);
	str.Add(m_strMIME,mlen);
	str.Add(buffer,strlen(buffer));
	str.Add("?=",2);

	while (jisleft(temp,42,jisbuf)) {
		m_encode.EncodeLine(strlen(jisbuf),jisbuf,buffer,sizeof(buffer));
		str.Add("\015\012 ",3);
		str.Add(m_strMIME,mlen);
		str.Add(buffer,strlen(buffer));
		str.Add("?=",2);
	}

	temp = str.GetBuffer();
	str.Detach();

	return temp;

}


/*
	Function: RFC1468::ConvToFrom
	Description:
		Convert To or From text.
*/
char *CRFC1468::ConvToFrom(const char* in)
{
	BCStr out;	
	char jisbuf[1024],buffer[1024];
	const char* input = in;
	if (input[0] == '@')
		input++;
	int len = Sjis2Jis(input,&out);
	if (len == 0) {			// no dbcs ? 
		char* temp = out.GetBuffer();
		out.Detach();
		return temp;		// no conversion
	}

	char* temp = out.GetBuffer();
	BCStr addrstr;
	int fmt = 0,pos1 = 0 ;		// assume < format
	//2000.01.14 Support TO: @header
	if (in[0] == '@')
		goto skip;
	
	pos1 = jisinchar(temp,'<');
	if (pos1 < 0) {			// maybe xxx@xxx.com (xxxx)  format
		fmt = 1;		// ( format
		char *tp = Strchr(temp,'(');
		char *tp2 = Strchr(temp,')');
		if (!tp || !tp2) {	// check ( and )
			out.Detach();
			return temp;
		} else {
			*tp = 0;
			addrstr.Add(temp,strlen(temp));	// real addr
			temp = tp + 1;
			*tp2 = 0;
		}
	} else {
		addrstr.Add(temp + pos1,strlen(temp+pos1));	//real addr
		temp[pos1] = 0;
	}

skip:

	int clen = jisleft(temp,36,jisbuf);	// 
	m_encode.EncodeLine(strlen(jisbuf),jisbuf,buffer,sizeof(buffer));
	BCStr str;
	int mlen = strlen(m_strMIME);
	str.Add(m_strMIME,mlen);
	str.Add(buffer,strlen(buffer));
	str.Add("?=",2);
	// do not insert CR LF 
	while (jisleft(temp,42,jisbuf)) {
		m_encode.EncodeLine(strlen(jisbuf),jisbuf,buffer,sizeof(buffer));
		str.Add("\015\012  ",4);
		str.Add(m_strMIME,mlen);
		str.Add(buffer,strlen(buffer));
		str.Add("?=",2);
	}
	if (fmt == 0) { // xxxxx<addr> ?
		//2000.01.14 Support TO: @header
		if (addrstr.GetLength() > 0)
			str.Add(addrstr.GetBuffer(),addrstr.GetLength());
		temp = str.GetBuffer();
		str.Detach();
		return temp;
	}
	// addr (xxxxxxx) format
	addrstr.Add("(",1);
	addrstr.Add(str.GetBuffer(),str.GetLength());
	addrstr.Add(")",1);
	temp = addrstr.GetBuffer();
	addrstr.Detach();

	return temp;

}


/*
	Function: RFC1468::ConvBody
	Description:
		Convert Body for jis.
*/
char* CRFC1468::ConvBody(const char* in,char *content_type)
{

	BCStr out,ret;
	ret.Add(MIME_HEADER,strlen(MIME_HEADER));
	if (content_type && strstr(content_type,"multipart")) { // 2000.10.29 multipart support
		ret.Add(content_type,strlen(content_type));
		ret.Add("\r\n\r\n",4);
		ret.Add((char*)in,strlen(in));
		char *temp = ret.GetBuffer();
		ret.Detach();
		return temp;
	}

	
	int len = Sjis2Jis(in,&out);
	// 1999.3.5 Content-Type support
	if (len == 0) { 			// no dbcs ? 
		// 1999.09.27 Content-Type support
		if (content_type) {
			char temp[256];
			sprintf(temp,MIME_BODY_WITHCONTENT,m_strContentType,m_strCharset); // 2000.09.14
			ret.Add(temp,strlen(temp));
		} else 
			ret.Add(MIME_BODY,strlen(MIME_BODY));
	} else {
		// 1999.08.26 Support other charset
		if (m_nCharset != 0) {
			char temp[256];
			sprintf(temp,MIME_BODY_OTHER,m_strContentType,m_strCharset); // 2000.09.14
			ret.Add(temp,strlen(temp));
			// convert body to base64
			out.Clear();
			char buffer[128];  // chage to base64
			int newlen = 0;
			for ( int i = 0;i < len;i += 51) {
				if (len - i >= 51)
					m_encode.EncodeLine(51,(char*)in+i,buffer,sizeof(buffer));
				else
					m_encode.EncodeLine(len - i,(char*)in+i,buffer,sizeof(buffer));
				out.Add(buffer,strlen(buffer));
				out.Add("\015\012",2);
				newlen += strlen(buffer) + 2;
			}
		} else {
			// 1999.09.27 Content-Type support
			if (content_type) {
				char temp[256];
			    sprintf(temp,MIME_BODY_WITHCONTENT,m_strContentType,m_strCharset); // 2000.09.14
				ret.Add(temp,strlen(temp));
			}
			else 
				ret.Add(MIME_BODY_JP,strlen(MIME_BODY_JP));
		}
	}

	BCStr out2;
	ConvOneDot(out,out2);	// convert "\n.\n"to "\n..\n" for smtp
	
	ret.Add(out2.GetBuffer(),out2.GetLength());
	
	char *temp = ret.GetBuffer();
	ret.Detach();

	return temp;

}
// 2000.05.17 Support varying Content-Type
static char* GetContentType(char *name)
{
	int fnamel = strlen(name);
	char *ext = name+fnamel -4;
	if (strcasecmp(ext,".txt") == 0)
		return "text/plain";
	if (strcasecmp(ext-1,".html") == 0 || 
		strcasecmp(ext,".htm") == 0) 
		return "text/html";
	if (strcasecmp(ext-1,".jpeg") == 0 || 
		strcasecmp(ext,".jpg") == 0) 
		return "image/jpeg";
	if (strcasecmp(ext,".gif") == 0)
		return "image/gif";
	if (strcasecmp(ext,".bmp") == 0)
		return "image/bmp";
	if (strcasecmp(ext,".zip") == 0)
		return "application/x-zip-compressed";
	if (strcasecmp(ext,".doc") == 0)
		return "application/msword";
	if (strcasecmp(ext,".xls") == 0)
		return "application/vnd.ms-excel";
	if (strcasecmp(ext,".ppt") == 0)
		return "application/vnd.ms-powerpoint";
	if (strcasecmp(ext,".exe") == 0 || strcasecmp(ext,".dll") == 0)
		return "application/x-msdownload";
	if (strcasecmp(ext,".eml") == 0 || strcasecmp(ext,".mht") == 0 || // 2000.08.01 Support message/rfc822
		strcasecmp(ext-1,".mhtml") == 0 || strcasecmp(ext,".nws") == 0) //  eml,mht,mhtml,nws
		return "message/rfc822";
	if (strcasecmp(ext,".fdf") == 0) // 2000.08.06 Support application/vnd.fdf
		return "application/vnd.fdf";
	if (strcasecmp(ext,".pdf") == 0) // 2000.08.06 Support application/pdf
		return "application/pdf";
	if (strcasecmp(ext,".rmi") == 0) // 2000.08.06 Support audio/mid
		return "audio/mid";
	if (strcasecmp(ext+1,".qt") == 0 || strcasecmp(ext,".mov") == 0) // 2000.08.06 video/quicktime
		return "video/quicktime";
	if (strcasecmp(ext,".avi") == 0) // 2000.08.06 Support video/x-msvideo
		return "video/x-msvideo";
	if (strcasecmp(ext-1,".tiff") == 0) // 2000.08.06 Support image/tiff
		return "image/tiff";
	if (strcasecmp(ext,".wav") == 0 || strcasecmp(ext,".mp1") == 0 || // 2000.08.06 Support audio/wav
		strcasecmp(ext,".mp2") == 0 || strcasecmp(ext,".mp3") == 0) 
		return "audio/wav";
	if (strcasecmp(ext,".als") == 0) // 2000.08.26 Support Audio/X-Alpha5
		return "Audio/X-Alpha5";

	return "application/octet-stream";
}


void CRFC1468::PutBase64(char *fnames,FILE *fp)
{

	// 2000.01.08 nick name support
	BCStr s;
	s.Add(fnames,strlen(fnames));
	char *fname = s.GetBuffer();
	char *tp = s.Find('|');
	// 2000.08.06 Support dynamic Content-Type
	char *type = NULL;
	// 2000.11.05 Support no encode
	char *noencode = NULL;
	if (tp) {
		type = Strchr(tp,'|');
		if (type) { // type specify
			type[0] = 0;
			type++;
			// 2000.11.05 noencode
			noencode = strchr(type,'|');
			if (noencode)
				noencode[0] = 0;
		}
		tp--;
	}
	
	if (tp == NULL || tp[1] == 0) { // 2000.08.06 Support Dynamic Content-Type
		tp = Strrchr(fname,'\\');
		if (tp == NULL)
			tp = Strrchr(fname,'/');
		if (tp == NULL)
			tp = fname-1;
	}

	char *name = ConvSubject(tp+1);	// name= para

	char buf[2048];
	// 1999.08.19 support text/html 
	// 2000.05.17 support varying Content-Type 
	if (type == NULL) // 2000.08.06 Support Dynamic Content-Type
		type = GetContentType(name);

	// 2000.08.01 support message/rfc822
	if (strcasecmp(type,"message/rfc822") != 0 && !noencode) {
		sprintf(buf,MIME_FILEPART,m_strBoundary,type,name); //2000.08.01 boundary
		fwrite(buf,1,strlen(buf),fp);
		if (m_encode.OpenFile(fname) != 0)
			return;
		while (m_encode.EncodeLine(buf,51) == 0) {	// length 51 for 68 bytes
			strcat(buf,"\015\012");
			fwrite(buf,1,strlen(buf),fp);
		}
		m_encode.CloseFile(); // 2000.08.01
	} else { // 2000.08.01 "message/rfc822" so write no conversion data
		sprintf(buf,MIME_FILEPART_7BIT,m_strBoundary,type,name);
		fwrite(buf,1,strlen(buf),fp);
		FILE *h = fopen(fname,"rb");
		if (h == NULL)
			return;
		size_t iNumBytes;
		while(iNumBytes = fread(buf,1,sizeof(buf),h)) {
			if (iNumBytes <= 0)
				break;
			fwrite(buf,1,iNumBytes,fp);
		}
		fclose(h);
	}

	delete []name;

}


// convert MultiPart
char* CRFC1468::ConvMulti(const char* body,const char* file,char *content_type)
{

	BCStr out,text;
	char buff[2048];
	// 1999.10.29 FIX NULL body cause abend !!!!!
	char *bodydata = (char*)body;
	if (!bodydata)
		bodydata = "\r\n";

	// 1999.08.21 FIX body is not sent when content-type is header specified
	int dbcs = Sjis2Jis(bodydata,&out);
	text.Add(MIME_HEADER,strlen(MIME_HEADER));
	if (content_type) {	// 1999.3.5 content-type support
		char temp[1024];
		// 1999.08.26 Support other charset
		if (m_nCharset != 0) {
			sprintf(temp,MIME_BODY_MULTI2,m_strBoundary,
				m_strBoundary,content_type,"base64");
			// convert body to base64
			out.Clear();
			char buffer[128];  // chage to base64
			int newlen = 0;
			int len = strlen(bodydata);
			for ( int i = 0;i < len;i += 51) {
				if (len - i >= 51)
					m_encode.EncodeLine(51,(char*)bodydata+i,buffer,sizeof(buffer));
				else
					m_encode.EncodeLine(len - i,(char*)bodydata+i,buffer,sizeof(buffer));
				out.Add(buffer,strlen(buffer));
				out.Add("\015\012",2);
				newlen += strlen(buffer) + 2;
			}
		} else {
			sprintf(temp,MIME_BODY_MULTI2,m_strBoundary,
				m_strBoundary,content_type,"7bit");
		}

		text.Add(temp,strlen(temp));
	}
	else {
		if (0 == dbcs) {	// no dbcs ?
			sprintf(buff,MIME_BODY_MULTI,m_strBoundary,m_strBoundary);
		}
		else {
			sprintf(buff,MIME_BODY_MULTI_JP,m_strBoundary,m_strBoundary);
		}
		text.Add(buff,strlen(buff));
	}

	BCStr out2;
	ConvOneDot(out,out2);	// convert "\n.\n"to "\n..\n" for smtp

	text.Add(out2.GetBuffer(),out2.GetLength());

	// 1999.10.04 check body ends CRLF
	char *last2 = text.GetBuffer() + text.GetLength() -2;
	if (memcmp(last2, CRLF,2) != 0)
		text.Add(CRLF,2);

	char newfile[L_tmpnam];
	if (NULL == tmpnam(newfile))
		newfile[0] = 0;

	FILE *pfile =  fopen(newfile,"wb");
	if (pfile == NULL){
		return NULL;
	}
	fwrite(text.GetBuffer(),1,text.GetLength(),pfile);
	
	text.Clear();
	text.Add((char*)file,strlen(file));

	char *tp = text.GetBuffer();
	// each file name split by TAB character
	int pos = 0;
	for (register int i = 0;tp[i] != 0;i++){
		if (tp[i] == '\t') {	// find next to char 
			tp[i] = 0;
			PutBase64(tp+pos,pfile);
			pos = i+1;
		}
	}
	PutBase64(tp+pos,pfile);

	sprintf(buff,MIME_BOUNDARY_END,m_strBoundary); // 2000.08.01 boundary
	fwrite(buff,1,strlen(buff),pfile);
	fclose(pfile);

	BCStr ret;
	ret.Add(newfile,strlen(newfile));
	
	char *temp = ret.GetBuffer();
	ret.Detach();

	return temp;		// return file name

}




/*
	Function: RFC1468::Sjis2Jis
	Description:
		Convert string SJIS to JIS.

    SJIS kanji code is 8 bit
    JIS kanji code is 7 bit
*/
int CRFC1468::Sjis2Jis(const char* str,BCStr* out)
{
	// 1999.08.26 Support other charset
	if (m_nCharset != 0) {
		unsigned char ch1;
		int len = strlen(str);
		for (register int i = 0;; i++) {
			ch1 = *(str+i);
			if (ch1 == 0)
				break;
			if (ch1 > 127) {  // not ascii include  ?
				out->Add((char*)str,len);
				return len;
			}
		}
		out->Add((char*)str,len);
		return 0;
	}
	
	int dbcs,kctr;
	dbcs = kctr = 0;
	register unsigned char ch1; 
	unsigned int zenkaku;
	int han;
	for (register int i = 0;; i++) {
		ch1 = *(str+i);
		if (ch1 == 0)
			break;
		if (iskanji(ch1)) {  // SJIS kanji start ?
			if(!dbcs) {
				dbcs = 1;
			    out->Add(TOJIS,3);    // put Esc sequence
			}
			// convert 1 kanji to jis
			unsigned short sjis = (ch1 << 8) | (*(str + ++i) & 0x00ff);
			unsigned short jis = sjis2jis(sjis);
			if (jis == 0)
				jis = 8739;
			out->AddByte(jis >> 8);
			out->AddByte(jis);
			kctr++;
			continue;
		}
		han = han2zen(&zenkaku, (unsigned char*)str + i);
		if (han == 0) {	// no hankaku
			if (dbcs) {  // kanji mode ?
				dbcs = 0;
			    out->Add(TOASCII,3);    // put Esc sequence
			}
			if (ch1 > 127)	// not 7 bit
				ch1 = 32;
			
			out->AddByte(ch1);
			continue;
		}
		if(!dbcs) {
			dbcs = 1;
		    out->Add(TOJIS,3);    // put Esc sequence
		}
		out->AddByte(zenkaku >> 8);
		out->AddByte(zenkaku);
		kctr++;
		i = i + han -1;		//擪2EJȋg?
	}
	if (dbcs) {  // kanji mode ?
	    out->Add(TOASCII,3);    // put Esc sequence
	}
	
	return kctr;
}
/*
Description
	Returns a specified number of characters from the left side of a string.
*/
int CRFC1468::jisleft(char* str,int needlen,char* out)
{
	int outctr,dbcs;
	outctr = dbcs = 0;
	register int need = needlen;
	register int i;
	for (i = 0; need > 0;) {
		if (*(str+i) == 0)
			break;
		if (memcmp(str+i,TOJIS,3) == 0) {  // jis start ?
			if (need < 8)	// enough room ? 
				break;
			dbcs = 1;
		    memcpy(out+outctr,TOJIS,3);    // put Esc sequence
		    memcpy(out+outctr+3,str+i+3,2);    // kani copy
			outctr += 5; i += 5; need -= 5;
			continue;
		}
		if (memcmp(str+i,TOASCII,3) == 0) {  // ASCII start ?
			dbcs = 0;
		    memcpy(out+outctr,TOASCII,3);    // put Esc sequence
			outctr += 3; i += 3; need -= 3;
			continue;
		}
		if (dbcs) {  // jis mode ?
			if (need < 5)	// enough room ?
				break;
			memcpy(out+outctr,str+i,2);    // copy jis
			outctr += 2; i += 2; need -= 2;
		} else {
		    *(out+outctr) = *(str+i);    // copy ASCII
			outctr++; i++; need--;
		}
	}
	if (dbcs) {  // in jis ?
	    memcpy(out+outctr,TOASCII,3);    // put Esc sequence
		outctr += 3; need -= 3;
	    memcpy(str,TOJIS,3);    // put Esc sequence
	    memcpy(str+3,str+i,strlen(str)-i+1);   // copy remain data
	} else {
	    memcpy(str,str+i,strlen(str)-i+1);   // copy remain data
	}
	*(out+outctr) = 0;
	
	return outctr;
}

/*
Description

	Returns the position of the first occurrence of one character
					within another.
*/
int CRFC1468::jisinchar(const char* str,char findchar)
{
	int dbcs = 0;
	for (register int i = 0;;) {
		if (*(str+i) == 0)
			break;
		if (memcmp(str+i,TOJIS,3) == 0) {  // jis start ?
			dbcs = 1;
			i += 3;
			continue;
		}
		if (memcmp(str+i,TOASCII,3) == 0) {  // ASCII start ?
			dbcs = 0;
			i += 3;
			continue;
		}
		if (dbcs) {  // jis mode ?
			i += 2; continue;
		}
		if (*(str+i) == findchar)
				return i;
		i++;
	}
	return -1;
}

// 1999.08.26 Support other charset
void CRFC1468::SetCharset(const char* str)
{
	char rname[128],*tp;
	char *name = strstr(str,"charset="); // 2000.09.14 FIX charset
	if (name)
		name = name + 8;
	else
		goto skip;
	if (name[0] == '"') {	// trim "
		strcpy(rname,name+1);
		rname[strlen(rname)-1] = 0;
	} else
		strcpy(rname,name);

	
	strcpy(m_strCharset,rname);
	
	strcpy(m_strMIME,"=?");
	strcat(m_strMIME,rname);
	strcat(m_strMIME,"?B?");

	if (strcasecmp(rname, "iso-2022-jp") != 0)
		m_nCharset = 1;		// set other

	// 2000.09.14 get content-type
skip:
	strcpy(rname,str);
	tp = strchr(rname,';');
	if (tp)
		tp[0] = 0;
	tp = strrchr(rname,' ');
	if (tp == NULL)
		tp = strchr(rname,':');
	if (tp && strlen(tp+1) < sizeof(m_strContentType))
		strcpy(m_strContentType,tp+1);

}
